#include "include/mns.h"
#include "include/inetraw.h"

/* function prototypes */
int udpscan(int *);
int synscan(int *, hdrinfo);
int finscan(int *, hdrinfo);
struct in_addr get_local_ip(void);

/* global variables */
time_t timedelay;

int stealthscan(struct in_addr target, FILE *fsp,
                int **plist, struct in_addr *hostaddress)
{
   int (*scanfunc)(int *, hdrinfo);
   hdrinfo s_ack;
   struct tcphdr tcp;
   struct ip ip;
   int i, j, k;
   int *pschkd;
   time_t timeout;
   long ack_diff = 0;
   char buffer[80];
   
   srand(time(NULL));

   pschkd = (int *)calloc(services * sizeof(int), 1);
   *plist = NULL;

   init_packet_devices(IPPROTO_TCP);

   target_ip = target;

   local_ip = get_local_ip();
   
   if(!local_ip.s_addr) {
      term_packet_devices();
      fprintf(fsp, "\n\t- no response from target machine");

      free(pschkd);
      return(0);
   } else *hostaddress = local_ip;

   /* this is a TH_SYN | TH_FIN check that happens to be are really useful
    * check.  If an ACK of +2 is received then we know almost for sure
    * that we're looking at a Windows based system.
    * If we get no response, we either hit an open port on a *NIX based
    * machine or the machine is firewalled.
    * If we get ACK +2 or hit a firewalled machine, FIN scan won't work.
    * As far as I know, I'm the first one who has come up with this test.
    */
    
   s_ack.ack = 0;
   src_port = htons((rand() % 10000) + 5000);
   s_ack.seq = rand();

   sendtcp(&s_ack, TH_SYN | TH_FIN);

   for(i = j = 0; i < 2 && !j; i++) { /* we try twice just to make sure */
      timeout = time(NULL); /* we also calibrate the time delay */
      while(timeout + TIMEOUT > time(NULL))
         if(get_packet(&ip, (char *)&tcp, sizeof(tcp), SRC))
            if(src_port == tcp.th_dport) { /* response to us? */
               if(ntohl(tcp.th_ack)) /* did we get an ACK? */
                  ack_diff = ntohl(tcp.th_ack) - s_ack.seq; /* ACK-SEQ */
               j = 1; /* yes, we got a valid packet */
               break;
            }
   }
   
   timedelay = time(NULL) - timeout + 4; /* calibration for portscan */

   if (ack_diff == 2 || !j) { /* is our response weird? */
      fprintf(fsp, "\n\t- non-standard response, reverting to SYN scan");
      fprintf(fsp, "\n\t- target likely to be a Windows based machine");
      scanfunc = synscan; /* weird tagets get SYN scaned -- more stable */
   } else scanfunc = finscan; /* go for the FIN cause our target is ok */

   if(mns_opts & SYNSCAN) scanfunc = synscan; /* forced SYN scan? */

   /* if we get too few open ports using a FIN scan, our target is suspect
    * of being extra-weird, so let's SYN them. examples of this are IRIX
    * and HP-UX but there are a few others
    */

   if((j = scanfunc(pschkd, s_ack)) < 2)
      if(scanfunc == finscan) {
         fprintf(fsp, "\n\t- FIN scan may have failed, trying SYN scan");
         j = synscan(pschkd, s_ack);
      }
   fprintf(fsp, "\n\tStandard TCP service ports open ->");
   if(j) {
      *plist = malloc((j + 1) * sizeof(int));
      (*plist)[0] = j;

      for (i = k = 0; i < services; i++) {
         if (!pschkd[i]) {
            k++;
	    (*plist)[k] = servlist[i].port;
            sprintf(buffer, "\n\t\t%3i  tcp\t%s",
                    servlist[i].port, servlist[i].name);
            fprintf(fsp, buffer);
         }
      }
   } else fprintf(fsp, " ( none )");

   free(pschkd);
   term_packet_devices();
   return(0);
}

int synscan(int *pschkd, hdrinfo s_ack) {
  struct tcphdr tcp;
  struct ip ip;
  int i, j, found = 0;
  time_t timeout;

  for (j = 0; j < services; j++)
    pschkd[j] = PORT_CLOSED;

   for (i = 0; i < services; i++) {
      dst_port = htons(servlist[i].port);

      s_ack.seq = rand();
      sendtcp(&s_ack, TH_SYN);

      if(get_packet(&ip, (char *)&tcp, sizeof(tcp), SRC))
         for (j = 0; j < services; j++)
            if(servlist[j].port == ntohs(tcp.th_sport)) {
               if(ntohs(tcp.th_win)) {
                  sendtcp(&s_ack, TH_RST);
                  if(pschkd[j] == PORT_CLOSED) {
#ifdef DEBUG
                     printf("port open %i : window %04x\n",
                            servlist[j].port,
                            ntohs(tcp.th_win));
#endif
                     pschkd[j] = 0;
                     found++;
                     break;
                  }
               }
            }
   }

   timeout = time (NULL) + timedelay;

   while (timeout > time (NULL)) {
      if (get_packet(&ip, (char *)&tcp, sizeof(tcp), SRC)) {
         for (j = 0; j < services; j++)
            if (servlist[j].port == ntohs (tcp.th_sport)) {
               if (ntohs (tcp.th_win)) {
                  sendtcp(&s_ack, TH_RST);
                  if(pschkd[j] == PORT_CLOSED) {
#ifdef DEBUG
                     printf("port open %i : window %04x\n",
                            servlist[j].port,
                            ntohs(tcp.th_win));
#endif
                     pschkd[j] = 0;
                     found++;
                     break;
                  }
               }
            }
      }
   }
   return found;
}

int finscan(int *pschkd, hdrinfo s_ack) {
   struct tcphdr tcp;
   struct ip ip;
   int i, j, found = services;
   time_t timeout;
   u_short tcpflags = 0;

   for (i = 0; i < services; i++) {
      dst_port = htons(servlist[i].port);
      
      s_ack.seq = rand();
      
      switch (i % 3) { /* we actually use NUL/FIN/PSH scanning */
	case 0: tcpflags = 0x0;     break; /* NULL */
	case 1: tcpflags = TH_FIN;  break; /* FIN  */
	case 2: tcpflags = TH_PUSH; break; /* PUSH */
      }

      sendtcp(&s_ack, tcpflags);

      if (get_packet(&ip, (char *)&tcp, sizeof(tcp), SRC)) {
         for(j = 0; j < services; j++)
            if(servlist[j].port == ntohs(tcp.th_sport)) {
               if(pschkd[j] != PORT_CLOSED) {
                  pschkd[j] = PORT_CLOSED;
                  found--;
                  break;
               }
            }
      }
   }
   timeout = time (NULL) + timedelay;

   while (timeout > time (NULL)) {
      if(get_packet(&ip, (char *)&tcp, sizeof(tcp), SRC)) {
         for (j = 0; j < services; j++)
            if(servlist[j].port == ntohs(tcp.th_sport)) {
               if(pschkd[j] != PORT_CLOSED) {
                  pschkd[j] = PORT_CLOSED;
                  found--;
                  break;
               }
            }
      }
   }

   return found;
}

struct in_addr get_local_ip(void) {
   SAin sa_dest;
   struct in_addr ip_addr;
   struct icmp_h icmp;
   struct ip ip;
   time_t timeout;

   ip_addr.s_addr = 0;

   bzero(&icmp, sizeof(icmp));
   icmp.icmp_type = ICMP_ECHO;
   icmp.icmp_un.echo.id = getpid();
   icmp.icmp_sum = in_cksum((u_short *)&icmp, sizeof(icmp));

   sa_dest.sin_family = AF_INET;
   sa_dest.sin_addr = target_ip;
   sa_dest.sin_port = htons((rand() % 1000) + 1000);
   
   sendto(sd_icmp, &icmp, sizeof(icmp), 0, (SA *)&sa_dest, sizeof(SAin));
   bzero(&icmp, sizeof(icmp));

   timeout = time(NULL) + TIMEOUT;
   while(timeout > time(NULL))
      if(get_packet(&ip, (char *)&icmp, sizeof(icmp), DST))
         if(icmp.icmp_type == ICMP_ECHO)
            if(icmp.icmp_un.echo.id == getpid()) {
               ip_addr = ip.ip_src;
               break;
            }

   return ip_addr;
}

#define udps_icmp ((struct icmp_h *)buffer)
#define udps_ip ((struct ip *)&buffer[sizeof(struct icmp_h)])
#define udps_udp ((struct udp_h *)&buffer[sizeof(struct icmp_h) +\
                                          sizeof(struct ip)])
int udpscan(int *pschkd) {
   struct ip ip;
   char buffer[sizeof(struct ip) +
               sizeof(struct udp_h) +
               sizeof(struct icmp_h)];
   int i, j, found = services;
   time_t timeout;

   for (i = 0; i < services; i++) {
      dst_port = htons(servlist[i].port);
      
      sendudp(sd_raw);

      if(!get_packet(&ip, buffer, sizeof(buffer), SRC)) continue;
      if(udps_icmp->icmp_type != ICMP_DEST_UNREACH) continue;
      if(udps_icmp->icmp_code != ICMP_PORT_UNREACH) continue;
      for(j = 0; j < services; j++)
         if(servlist[j].port == ntohs(udps_udp->udp_dport)) {
            pschkd[j] = PORT_CLOSED;
            found--;
            break;
         }

   }
   
   timeout = time(NULL) + TIMEOUT;

   while (timeout > time (NULL))
      if (get_packet(&ip, buffer, sizeof(buffer), SRC))
         if(udps_icmp->icmp_type == ICMP_DEST_UNREACH)
            if(udps_icmp->icmp_code == ICMP_PORT_UNREACH)
               for(j = 0; j < services; j++)
                  if(servlist[j].port == ntohs(udps_udp->udp_dport)) {
                     pschkd[j] = PORT_CLOSED;
                     found--;
                     break;
                  }
   return found;
}

int udpportscan(struct in_addr target, FILE *fsp, struct in_addr hostaddress)
{
   int i, j, k;
   int *pschkd;
   char buffer[80];
   srand(time(NULL));

   pschkd = (int *)calloc(services * sizeof(int), 1);
   init_packet_devices(IPPROTO_UDP);

   target_ip = target;
   local_ip = hostaddress;

   j = udpscan(pschkd);
   if(services - j < 10) {
      j = 0;
      fprintf(fsp, "\n\t- UDP scan may have failed, "
                   "no results will be shown");
   }
   
   fprintf(fsp, "\n\tStandard UDP service ports open ->");
   if(j)
      for (i = k = 0; i < services; i++)
         if (!pschkd[i]) {
            k++;
            sprintf (buffer, "\n\t\t%3i  udp\t%s",
            servlist[i].port, servlist[i].name);
            fprintf(fsp, buffer);
         }
   else fprintf(fsp, " ( none )");

  free(pschkd);
  term_packet_devices();
  return(0);
}

